'use strict';

const resources = require('./resources');

const DEFAULT_HOST = 'payments.ccamatil.com';
const DEFAULT_TEST_HOST = 'payments.ccamatil.com';
const DEFAULT_PORT = '443';
const DEFAULT_BASE_PATH = '/api/v2/';
const DEFAULT_TEST_BASE_PATH = '/uat/api/';
const DEFAULT_API_VERSION = null;

const DEFAULT_TIMEOUT = 80000;

Cpp.PACKAGE_VERSION = require('../package.json').version;

const utils = require('./utils');
const {determineProcessUserAgentProperties, emitWarning} = utils;

Cpp.USER_AGENT = {
    bindings_version: Cpp.PACKAGE_VERSION,
    lang: 'node',
    publisher: 'cpp',
    uname: null,
    typescript: false,
    ...determineProcessUserAgentProperties(),
};

/** @private */
Cpp._UNAME_CACHE = null;

const MAX_NETWORK_RETRY_DELAY_SEC = 2;
const INITIAL_NETWORK_RETRY_DELAY_SEC = 0.5;

const APP_INFO_PROPERTIES = ['name', 'version', 'url', 'partner_id'];
const ALLOWED_CONFIG_PROPERTIES = [
  'apiVersion',
  'typescript',
  'maxNetworkRetries',
  'httpAgent',
  'httpClient',
  'timeout',
  'host',
  'port',
  'basePath',
  'protocol',
  'telemetry',
  //'appInfo',
  'cppAccount',
  'test'
];

const EventEmitter = require('events').EventEmitter;

Cpp.CppResource = require('./CppResource');
Cpp.resources = resources;

const {HttpClient, HttpClientResponse} = require('./net/HttpClient');
Cpp.HttpClient = HttpClient;
Cpp.HttpClientResponse = HttpClientResponse;

const CryptoProvider = require('./crypto/CryptoProvider');
Cpp.CryptoProvider = CryptoProvider;

function Cpp(merchantId, sharedSecret, config = {}) {
    if (!(this instanceof Cpp)) {
        return new Cpp(merchantId, sharedSecret, config);
    }

    const props = this._getPropsFromConfig(config);

    Object.defineProperty(this, '_emitter', {
      value: new EventEmitter(),
      enumerable: false,
      configurable: false,
      writable: false,
    });

    this.VERSION = Cpp.PACKAGE_VERSION;

    this.on = this._emitter.on.bind(this._emitter);
    this.once = this._emitter.once.bind(this._emitter);
    this.off = this._emitter.removeListener.bind(this._emitter);

    if 
    (
        props.protocol &&
        props.protocol !== 'https' &&
        (!props.host || /\.payments\.ccamatil\.com$/.test(props.host))
    ) 
    {
        throw new Error(
        'The `https` protocol must be used when sending requests to `*.ccaaudevcppapi.azurewebsites.net`'
        );
    }

    const agent = props.httpAgent || null;

    this._api = {
        auth: null,
        merchantId: null,
        sharedSecret: null,
        host: props.host || this._setHost(props.test) ||  DEFAULT_HOST,
        port: props.port || DEFAULT_PORT,
        protocol: props.protocol || 'https',
        basePath: props.basePath || this._setBasePath(props.test) || DEFAULT_BASE_PATH,
        version: props.apiVersion || DEFAULT_API_VERSION,
        timeout: utils.validateInteger('timeout', props.timeout, DEFAULT_TIMEOUT),
        maxNetworkRetries: utils.validateInteger(
          'maxNetworkRetries',
          props.maxNetworkRetries,
          0
        ),
        agent: agent,
        httpClient: props.httpClient || Cpp.createNodeHttpClient(agent),
        crypto: Cpp.createNodeCryptoProvider(),
        dev: false,
        cppAccount: props.cppAccount || null,
    };

    const typescript = props.typescript || false;
    if (typescript !== Cpp.USER_AGENT.typescript) {
        // The mutation here is uncomfortable, but likely fastest;
        // serializing the user agent involves shelling out to the system,
        // and given some users may instantiate the library many times without switching between TS and non-TS,
        // we only want to incur the performance hit when that actually happens.
        Cpp.USER_AGENT.typescript = typescript;
    }

    // if (props.appInfo) {
    //     this._setAppInfo(props.appInfo);
    // }

    this._prepResources();
    //this._setApiKey(key);
    this._setApiMerchantId(merchantId);
    this._setApiSharedSecret(sharedSecret);

    this.errors = require('./Error');
    //this.webhooks = require('./Webhooks');

    this._prevRequestMetrics = [];
    this._enableTelemetry = props.telemetry !== false;

    // Expose CppResource on the instance too
    this.CppResource = Cpp.CppResource;
}

Cpp.errors = require('./Error');
//Cpp.webhooks = require('./Webhooks');

Cpp.createNodeHttpClient = (agent) => {
  const {NodeHttpClient} = require('./net/NodeHttpClient');
  return new NodeHttpClient(agent);
};

Cpp.createNodeCryptoProvider = () => {
  const NodeCryptoProvider = require('./crypto/NodeCryptoProvider');
  return new NodeCryptoProvider();
};

Cpp.prototype = {
  /**
   * @private
   */
  _setApiKey(key) {
    if (key) {
      this._setApiField('auth', `Bearer ${key}`);
    }
  },

  _setApiMerchantId(merchantId){
    if (merchantId){
      this._setApiField('merchantId', merchantId);
    }
  },

  _setApiSharedSecret(sharedSecret){
    if (sharedSecret){
      this._setApiField('sharedSecret', sharedSecret);
    }
  },

  /**
   * @private
   * This may be removed in the future.
   */
   _setHost(test) {
    if(test){
       return DEFAULT_TEST_HOST;
    }
    else{
       return DEFAULT_HOST;
    }
  },

  _setBasePath(test){
    if(test){
      return DEFAULT_TEST_BASE_PATH;
    }
    else{
      return DEFAULT_BASE_PATH;
    }
  },
  
  // _setAppInfo(info) {
  //   if (info && typeof info !== 'object') {
  //     throw new Error('AppInfo must be an object.');
  //   }

  //   if (info && !info.name) {
  //     throw new Error('AppInfo.name is required');
  //   }

  //   info = info || {};

  //   const appInfo = APP_INFO_PROPERTIES.reduce((accum, prop) => {
  //     if (typeof info[prop] == 'string') {
  //       accum = accum || {};

  //       accum[prop] = info[prop];
  //     }

  //     return accum;
  //   }, undefined);

  //   this._appInfo = appInfo;
  // },

  /**
   * @private
   * This may be removed in the future.
   */
  _setApiField(key, value) {
    this._api[key] = value;
  },

  /**
   * @private
   */
  getApiField(key) {
    return this._api[key];
  },

  /**
   * @private
   */
  getConstant: (c) => {
    switch (c) {
      case 'DEFAULT_HOST':
        return DEFAULT_HOST;
      case 'DEFAULT_PORT':
        return DEFAULT_PORT;
      case 'DEFAULT_BASE_PATH':
        return DEFAULT_BASE_PATH;
      case 'DEFAULT_API_VERSION':
        return DEFAULT_API_VERSION;
      case 'DEFAULT_TIMEOUT':
        return DEFAULT_TIMEOUT;
      case 'MAX_NETWORK_RETRY_DELAY_SEC':
        return MAX_NETWORK_RETRY_DELAY_SEC;
      case 'INITIAL_NETWORK_RETRY_DELAY_SEC':
        return INITIAL_NETWORK_RETRY_DELAY_SEC;
    }
    return Cpp[c];
  },

  getMaxNetworkRetries() {
    return this.getApiField('maxNetworkRetries');
  },

  /**
   * @private
   * This may be removed in the future.
   */
  _setApiNumberField(prop, n, defaultVal) {
    const val = utils.validateInteger(prop, n, defaultVal);

    this._setApiField(prop, val);
  },

  getMaxNetworkRetryDelay() {
    return MAX_NETWORK_RETRY_DELAY_SEC;
  },

  getInitialNetworkRetryDelay() {
    return INITIAL_NETWORK_RETRY_DELAY_SEC;
  },

  /**
   * @private
   */
  getUname(cb) {
    if (!Cpp._UNAME_CACHE) {
        Cpp._UNAME_CACHE = new Promise((resolve) => {
        utils.safeExec('uname -a', (err, uname) => {
          resolve(uname);
        });
      });
    }
    Cpp._UNAME_CACHE.then((uname) => cb(uname));
  },

  /**
   * @private
   * Gets a JSON version of a User-Agent and uses a cached version for a slight
   * speed advantage.
   */
  getClientUserAgent(cb) {
    return this.getClientUserAgentSeeded(Cpp.USER_AGENT, cb);
  },

  /**
   * @private
   * Gets a JSON version of a User-Agent by encoding a seeded object and
   * fetching a uname from the system.
   */
  getClientUserAgentSeeded(seed, cb) {
    this.getUname((uname) => {
      const userAgent = {};
      for (const field in seed) {
        userAgent[field] = encodeURIComponent(seed[field]);
      }

      // URI-encode in case there are unusual characters in the system's uname.
      userAgent.uname = encodeURIComponent(uname || 'UNKNOWN');

      const client = this.getApiField('httpClient');
      if (client) {
        userAgent.httplib = encodeURIComponent(client.getClientName());
      }

      // if (this._appInfo) {
      //   userAgent.application = this._appInfo;
      // }

      cb(JSON.stringify(userAgent));
    });
  },

  /**
   * @private
   */
  // getAppInfoAsString() {
  //   if (!this._appInfo) {
  //     return '';
  //   }

  //   let formatted = this._appInfo.name;

  //   if (this._appInfo.version) {
  //     formatted += `/${this._appInfo.version}`;
  //   }

  //   if (this._appInfo.url) {
  //     formatted += ` (${this._appInfo.url})`;
  //   }

  //   return formatted;
  // },

  getTelemetryEnabled() {
    return this._enableTelemetry;
  },

  /**
   * @private
   * This may be removed in the future.
   */
  _prepResources() {
    for (const name in resources) {
      this[utils.pascalToCamelCase(name)] = new resources[name](this);
    }
  },

  /**
   * @private
   * This may be removed in the future.
   */
  _getPropsFromConfig(config) {
    // If config is null or undefined, just bail early with no props
    if (!config) {
      return {};
    }

    // config can be an object or a string
    const isString = typeof config === 'string';
    const isObject = config === Object(config) && !Array.isArray(config);

    if (!isObject && !isString) {
      throw new Error('Config must either be an object or a string');
    }

    // If config is a string, we assume the old behavior of passing in a string representation of the api version
    if (isString) {
      return {
        apiVersion: config,
      };
    }

    // If config is an object, we assume the new behavior and make sure it doesn't contain any unexpected values
    const values = Object.keys(config).filter(
      (value) => !ALLOWED_CONFIG_PROPERTIES.includes(value)
    );

    if (values.length > 0) {
      throw new Error(
        `Config object may only contain the following: ${ALLOWED_CONFIG_PROPERTIES.join(
          ', '
        )}`
      );
    }

    return config;
  },
};

module.exports = Cpp;

// expose constructor as a named property to enable mocking with Sinon.JS
module.exports.Cpp = Cpp;

// Allow use with the TypeScript compiler without `esModuleInterop`.
// We may also want to add `Object.defineProperty(exports, "__esModule", {value: true});` in the future, so that Babel users will use the `default` version.
module.exports.default = Cpp;