'use strict';

const crypto = require('crypto');

const CryptoProvider = require('./CryptoProvider');
const util = require('util');

/**
 * `CryptoProvider which uses the Node `crypto` package for its computations.
 */
class NodeCryptoProvider extends CryptoProvider {

  /** @override */
  computeHMACSignature(payload, secret) {
    return crypto
      .createHmac('sha256', secret)
      .update(payload, 'utf8')
      .digest('hex');
  }

  /** @override */
  computeHMACSignatureBase64(payload, secret) {
    return crypto
      .createHmac('sha256', secret)
      .update(payload, 'utf8')
      .digest('base64');
  }

  /** @override */
  computeHashBase64(payload) {
    return crypto
      .createHash('sha256')
      .update(payload, 'utf8')
      .digest('base64');
  }

  
  /**
  * @override
  * Generates Hash and Signature via shared secret method based on payload and headers
  */
  generateSignedAuthHeader(merchantId, timestamp, payload, sharedSecret) {
    
    const hash = this.computeHashBase64(payload);

    const headerString = util.format('MerchantId: %s\nTimestamp: %s\nHash: %s', merchantId, timestamp, hash);
    
    const sign = this.computeHMACSignatureBase64(headerString, Buffer.from(sharedSecret, 'base64'));

    return {
        merchantId,
        hash,
        sign
    }
  }
}

module.exports = NodeCryptoProvider;
